@@ -31,6 +31,8 @@
 #include <fcntl.h>
 #include <uuid.h>
 
+#include <algorithm>
+
 #if defined(__FreeBSD__) || defined(__DragonFly__) || defined(__OpenBSD__)
 #include <sys/sysctl.h>
 #endif
@@ -62,7 +64,9 @@
 
 #include "eggshell.hh"
 
-#define URL_MATCH_CURSOR  (GDK_HAND2)
+#define URL_MATCH_CURSOR_NAME "pointer"
+
+namespace {
 
 typedef struct {
   volatile int refcount;
@@ -89,6 +93,8 @@ typedef struct {
   GCancellable *cancellable;
 } ExecData;
 
+} // anon namespace
+
 typedef struct
 {
   int tag;
@@ -108,6 +114,15 @@ struct _TerminalScreenPrivate
   gboolean exec_on_realize;
   guint idle_exec_source;
   ExecData *exec_data;
+
+  GIcon* icon_color;
+  GIcon* icon_image;
+
+  bool has_progress;
+  bool icon_progress_set;
+  VteProgressHint progress_hint;
+  double progress_fraction;
+  GIcon* icon_progress;
 };
 
 enum
@@ -120,11 +135,14 @@ enum
 };
 
 enum {
-  PROP_0,
-  PROP_PROFILE,
+  PROP_PROFILE = 1,
+  PROP_ICON,
   PROP_TITLE,
+  N_PROPS
 };
 
+static GParamSpec* pspecs[N_PROPS];
+
 enum
 {
   TARGET_COLOR,
@@ -414,6 +432,217 @@ terminal_screen_class_enable_menu_bar_accel_notify_cb (GSettings *settings,
     gtk_binding_entry_skip (binding_set, GDK_KEY_F10, GDK_SHIFT_MASK);
 }
 
+static void
+terminal_screen_icon_color_changed_cb(TerminalScreen* screen,
+                                      char const* prop,
+                                      VteTerminal* terminal)
+{
+  auto const priv = screen->priv;
+
+  g_clear_object(&priv->icon_color);
+
+  auto color = GdkRGBA{};
+  if (vte_terminal_get_termprop_rgba_by_id(terminal,
+                                           VTE_PROPERTY_ID_ICON_COLOR,
+                                           &color)) {
+    auto const scale = gtk_widget_get_scale_factor(GTK_WIDGET(screen));
+    auto const w = 32 * scale, h = 32 * scale;
+    auto const xc = w / 2, yc = h / 2;
+    auto const radius = w / 2 - 1;
+
+    auto surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, w, h);
+    auto cr = cairo_create(surface);
+    cairo_set_source_rgb(cr, color.red, color.green, color.blue);
+    cairo_new_sub_path(cr);
+    cairo_arc(cr, xc, yc, radius, 0., G_PI * 2);
+    cairo_close_path(cr);
+    cairo_fill(cr);
+
+    priv->icon_color =
+      G_ICON(gdk_pixbuf_get_from_surface(surface,
+                                         0,
+                                         0,
+                                         cairo_image_surface_get_width(surface),
+                                         cairo_image_surface_get_height(surface)));
+
+    cairo_destroy(cr);
+    cairo_surface_destroy(surface);
+  }
+
+  g_object_notify_by_pspec(G_OBJECT(terminal), pspecs[PROP_ICON]);
+}
+
+static void
+terminal_screen_icon_image_changed_cb(TerminalScreen* screen,
+                                      char const* prop,
+                                      VteTerminal* terminal)
+{
+  auto const priv = screen->priv;
+
+  g_clear_object(&priv->icon_image);
+  priv->icon_image =
+    G_ICON(vte_terminal_ref_termprop_image_pixbuf_by_id(terminal,
+                                                        VTE_PROPERTY_ID_ICON_IMAGE));
+
+  g_object_notify_by_pspec(G_OBJECT(terminal), pspecs[PROP_ICON]);
+}
+
+static GIcon*
+terminal_screen_ensure_icon_progress(TerminalScreen* screen)
+{
+  auto const priv = screen->priv;
+
+  if (priv->icon_progress_set)
+    return priv->icon_progress;
+
+  if (priv->has_progress) {
+    GIcon* icon = nullptr;
+
+    switch (priv->progress_hint) {
+    case VTE_PROGRESS_HINT_ERROR:
+      icon = g_themed_icon_new("dialog-error-symbolic");
+      break;
+
+    case VTE_PROGRESS_HINT_INDETERMINATE:
+      icon = nullptr;
+      break;
+
+    case VTE_PROGRESS_HINT_PAUSED:
+    case VTE_PROGRESS_HINT_ACTIVE: {
+      auto const scale = gtk_widget_get_scale_factor(GTK_WIDGET(screen));
+      auto const w = 16 * scale, h = 16 * scale;
+      auto const xc = w / 2, yc = h / 2;
+      auto const radius = w / 2 - 1;
+
+      auto color = GdkRGBA{};
+      G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
+      auto style_context = gtk_widget_get_style_context(GTK_WIDGET(screen));
+      gtk_style_context_get_color(style_context, gtk_style_context_get_state(style_context), &color);
+      G_GNUC_END_IGNORE_DEPRECATIONS;
+
+      auto surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, w, h);
+      auto cr = cairo_create(surface);
+
+      // First draw a shadow filled circle
+      cairo_set_source_rgba(cr, color.red, color.green, color.blue, 0.25);
+      cairo_arc(cr, xc, yc, radius, 0., 2 * G_PI);
+      cairo_close_path(cr);
+      cairo_fill(cr);
+
+      // Now draw progress filled circle
+      auto const fraction = priv->progress_fraction;
+      if (fraction > 0.) {
+        cairo_set_line_width(cr, 1.);
+        cairo_set_source_rgb(cr, color.red, color.green, color.blue);
+        cairo_new_sub_path(cr);
+
+        if (fraction < 1.) {
+          cairo_move_to(cr, xc, yc);
+          cairo_line_to(cr, xc + radius, yc);
+          cairo_arc_negative(cr, xc, yc, radius, 0, 2 * G_PI * (1. - fraction));
+          cairo_line_to(cr, xc, yc);
+        } else {
+          cairo_arc(cr, xc, yc, radius, 0, 2 * G_PI);
+        }
+
+        cairo_close_path(cr);
+        cairo_fill(cr);
+      }
+      cairo_destroy(cr);
+
+      icon = G_ICON(gdk_pixbuf_get_from_surface(surface,
+                                                0,
+                                                0,
+                                                cairo_image_surface_get_width(surface),
+                                                cairo_image_surface_get_height(surface)));
+
+      cairo_surface_destroy(surface);
+      break;
+    }
+
+    case VTE_PROGRESS_HINT_INACTIVE:
+    default:
+      icon = nullptr;
+      break;
+    }
+
+    g_set_object(&priv->icon_progress, icon);
+
+  } else {
+    // Remove progress
+    g_clear_object(&priv->icon_progress);
+  }
+
+  priv->icon_progress_set = true;
+
+  return priv->icon_progress;
+}
+
+static void
+terminal_screen_clear_icon_progress(TerminalScreen* screen)
+{
+  auto const priv = screen->priv;
+
+  g_clear_object(&priv->icon_progress);
+  priv->icon_progress_set = false;
+
+  g_object_notify_by_pspec(G_OBJECT(screen), pspecs[PROP_ICON]);
+}
+
+static void
+terminal_screen_progress_value_changed_cb(TerminalScreen* screen,
+                                          char const* prop,
+                                          VteTerminal* terminal)
+{
+  auto const priv = screen->priv;
+
+  auto fraction = 0.;
+  uint64_t value = 0;
+  auto const has_progress = vte_terminal_get_termprop_uint_by_id(terminal,
+                                                                 VTE_PROPERTY_ID_PROGRESS_VALUE,
+                                                                 &value);
+  if (has_progress)
+    fraction = std::clamp(double(value) / 100., 0., 1.);
+  else
+    fraction = 0.;
+
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wfloat-equal"
+  if (priv->has_progress == has_progress &&
+      priv->progress_fraction == fraction)
+    return;
+#pragma GCC diagnostic pop
+
+  priv->has_progress = has_progress;
+  priv->progress_fraction = fraction;
+
+  terminal_screen_clear_icon_progress(screen);
+}
+
+static void
+terminal_screen_progress_hint_changed_cb(TerminalScreen* screen,
+                                         char const* prop,
+                                         VteTerminal* terminal)
+{
+  auto const priv = screen->priv;
+
+  VteProgressHint hint;
+  int64_t value = 0;
+  if (vte_terminal_get_termprop_int_by_id(terminal,
+                                          VTE_PROPERTY_ID_PROGRESS_HINT,
+                                          &value))
+    hint = VteProgressHint(value);
+  else
+    hint = VTE_PROGRESS_HINT_INACTIVE;
+
+  if (priv->progress_hint == hint)
+    return;
+
+  priv->progress_hint = hint;
+
+  terminal_screen_clear_icon_progress(screen);
+}
+
 static TerminalWindow *
 terminal_screen_get_window (TerminalScreen *screen)
 {
@@ -491,6 +720,10 @@ terminal_screen_init (TerminalScreen *screen)
 
   priv->child_pid = -1;
 
+  priv->has_progress = false;
+  priv->progress_hint = VTE_PROGRESS_HINT_INACTIVE;
+  priv->progress_fraction = 0.;
+
   vte_terminal_set_allow_hyperlink (terminal, TRUE);
 
   for (i = 0; i < n_url_regexes; ++i)
@@ -500,7 +733,7 @@ terminal_screen_init (TerminalScreen *screen)
       tag_data = g_slice_new (TagData);
       tag_data->flavor = url_regex_flavors[i];
       tag_data->tag = vte_terminal_match_add_regex (terminal, url_regexes[i], 0);
-      vte_terminal_match_set_cursor_type (terminal, tag_data->tag, URL_MATCH_CURSOR);
+      vte_terminal_match_set_cursor_name (terminal, tag_data->tag, URL_MATCH_CURSOR_NAME);
 
       priv->match_tags = g_slist_prepend (priv->match_tags, tag_data);
     }
@@ -527,6 +760,16 @@ terminal_screen_init (TerminalScreen *screen)
                     G_CALLBACK (terminal_screen_window_title_changed),
                     screen);
 
+  g_signal_connect(screen, "termprop-changed::" VTE_TERMPROP_ICON_COLOR,
+                   G_CALLBACK(terminal_screen_icon_color_changed_cb), screen);
+  g_signal_connect(screen, "termprop-changed::" VTE_TERMPROP_ICON_IMAGE,
+                   G_CALLBACK(terminal_screen_icon_image_changed_cb), screen);
+
+  g_signal_connect(screen, "termprop-changed::" VTE_TERMPROP_PROGRESS_VALUE,
+                   G_CALLBACK(terminal_screen_progress_value_changed_cb), screen);
+  g_signal_connect(screen, "termprop-changed::" VTE_TERMPROP_PROGRESS_HINT,
+                   G_CALLBACK(terminal_screen_progress_hint_changed_cb), screen);
+
   app = terminal_app_get ();
   g_signal_connect (terminal_app_get_desktop_interface_settings (app), "changed::" MONOSPACE_FONT_KEY_NAME,
                     G_CALLBACK (terminal_screen_system_font_changed_cb), screen);
@@ -546,6 +789,9 @@ terminal_screen_get_property (GObject *object,
       case PROP_PROFILE:
         g_value_set_object (value, terminal_screen_get_profile (screen));
         break;
+      case PROP_ICON:
+        g_value_set_object(value, terminal_screen_get_icon(screen));
+        break;
       case PROP_TITLE:
         g_value_set_string (value, terminal_screen_get_title (screen));
         break;
@@ -568,6 +814,7 @@ terminal_screen_set_property (GObject *object,
       case PROP_PROFILE:
         terminal_screen_set_profile (screen, (GSettings*)g_value_get_object (value));
         break;
+      case PROP_ICON:
       case PROP_TITLE:
         /* not writable */
       default:
@@ -639,25 +886,29 @@ terminal_screen_class_init (TerminalScreenClass *klass)
                   G_TYPE_NONE,
                   0);
 
-  g_object_class_install_property
-    (object_class,
-     PROP_PROFILE,
+  pspecs[PROP_PROFILE] =
      g_param_spec_object ("profile", nullptr, nullptr,
                           G_TYPE_SETTINGS,
                           GParamFlags(G_PARAM_READWRITE |
 				      G_PARAM_STATIC_NAME |
 				      G_PARAM_STATIC_NICK |
-				      G_PARAM_STATIC_BLURB)));
+				      G_PARAM_STATIC_BLURB));
 
-  g_object_class_install_property
-    (object_class,
-     PROP_TITLE,
+  pspecs[PROP_ICON] =
+     g_param_spec_object ("icon", nullptr, nullptr,
+                          G_TYPE_ICON,
+                          GParamFlags(G_PARAM_READABLE |
+				      G_PARAM_STATIC_STRINGS |
+                                      G_PARAM_EXPLICIT_NOTIFY));
+
+  pspecs[PROP_TITLE] =
      g_param_spec_string ("title", nullptr, nullptr,
                           nullptr,
                           GParamFlags(G_PARAM_READABLE |
-				      G_PARAM_STATIC_NAME |
-				      G_PARAM_STATIC_NICK |
-				      G_PARAM_STATIC_BLURB)));
+				      G_PARAM_STATIC_STRINGS |
+                                      G_PARAM_EXPLICIT_NOTIFY));
+
+  g_object_class_install_properties(object_class, N_PROPS, pspecs);
 
   g_type_class_add_private (object_class, sizeof (TerminalScreenPrivate));
 
@@ -710,6 +961,9 @@ terminal_screen_dispose (GObject *object)
 
   terminal_screen_clear_exec_data (screen, TRUE);
 
+  g_clear_object(&screen->priv->icon_color);
+  g_clear_object(&screen->priv->icon_image);
+
   G_OBJECT_CLASS (terminal_screen_parent_class)->dispose (object);
 
   /* Unregister *after* chaining up to the parent's dispose,
@@ -1061,6 +1315,9 @@ terminal_screen_profile_changed_cb (GSettings     *profile,
   if (!prop_name || prop_name == I_(TERMINAL_PROFILE_AUDIBLE_BELL_KEY))
       vte_terminal_set_audible_bell (vte_terminal, g_settings_get_boolean (profile, TERMINAL_PROFILE_AUDIBLE_BELL_KEY));
 
+  if (!prop_name || prop_name == I_(TERMINAL_PROFILE_SCROLL_ON_INSERT_KEY))
+    vte_terminal_set_scroll_on_insert(vte_terminal,
+                                      g_settings_get_boolean(profile, TERMINAL_PROFILE_SCROLL_ON_INSERT_KEY));
   if (!prop_name || prop_name == I_(TERMINAL_PROFILE_SCROLL_ON_KEYSTROKE_KEY))
     vte_terminal_set_scroll_on_keystroke (vte_terminal,
                                           g_settings_get_boolean (profile, TERMINAL_PROFILE_SCROLL_ON_KEYSTROKE_KEY));
@@ -1280,7 +1537,7 @@ terminal_screen_set_profile (TerminalScreen *screen,
   if (old_profile)
     g_object_unref (old_profile);
 
-  g_object_notify (G_OBJECT (screen), "profile");
+  g_object_notify_by_pspec (G_OBJECT (screen), pspecs[PROP_PROFILE]);
 }
 
 GSettings*
@@ -1407,6 +1664,20 @@ terminal_screen_get_child_command (TerminalScreen *screen,
   return TRUE;
 }
 
+static gboolean
+remove_prefixed_cb(void* key,
+                   void* value,
+                   void* user_data)
+{
+  auto const env = reinterpret_cast<char const*>(key);
+  auto const prefix = reinterpret_cast<char const*>(user_data);
+
+  if (terminal_client_get_environment_prefix_filters_is_excluded(env))
+    return false;
+
+  return g_str_has_prefix(env, prefix);
+}
+
 static char**
 terminal_screen_get_child_environment (TerminalScreen *screen,
                                        char **initial_envv,
@@ -1442,10 +1713,17 @@ terminal_screen_get_child_environment (TerminalScreen *screen,
     }
 
   /* Remove unwanted env variables */
-  char const* const* filters = terminal_client_get_environment_filters ();
+  auto const filters = terminal_client_get_environment_filters ();
   for (i = 0; filters[i]; ++i)
     g_hash_table_remove (env_table, filters[i]);
 
+  auto const pfilters = terminal_client_get_environment_prefix_filters ();
+  for (i = 0; pfilters[i]; ++i) {
+    g_hash_table_foreach_remove (env_table,
+                                 GHRFunc(remove_prefixed_cb),
+                                 (void*)pfilters[i]);
+  }
+
   terminal_util_add_proxy_env (env_table);
 
   /* Add gnome-terminal private env vars used to communicate back to g-t-server */
@@ -1494,7 +1772,8 @@ info_bar_response_cb (GtkWidget *info_bar,
     case RESPONSE_EDIT_PREFERENCES:
       terminal_app_edit_preferences (terminal_app_get (),
                                      terminal_screen_get_profile (screen),
-                                     "custom-command-entry");
+                                     "custom-command-entry",
+                                     gtk_get_current_event_time());
       break;
     default:
       gtk_widget_destroy (info_bar);
@@ -1824,7 +2103,7 @@ static void
 terminal_screen_window_title_changed (VteTerminal *vte_terminal,
                                       TerminalScreen *screen)
 {
-  g_object_notify (G_OBJECT (screen), "title");
+  g_object_notify_by_pspec (G_OBJECT (screen), pspecs[PROP_TITLE]);
 }
 
 static void
@@ -2248,7 +2527,7 @@ terminal_screen_has_foreground_process (TerminalScreen *screen,
   if (sysctl (mib, G_N_ELEMENTS (mib), nullptr, &len, nullptr, 0) == -1)
       return TRUE;
 
-  data_buf = g_malloc0 (len);
+  data_buf = (char*)g_malloc0 (len);
   if (sysctl (mib, G_N_ELEMENTS (mib), data_buf, &len, nullptr, 0) == -1)
       return TRUE;
   data = data_buf;
@@ -2260,7 +2539,7 @@ terminal_screen_has_foreground_process (TerminalScreen *screen,
   if (sysctl (mib, G_N_ELEMENTS (mib), nullptr, &len, nullptr, 0) == -1)
       return TRUE;
 
-  data_buf = g_malloc0 (len);
+  data_buf = (char*)g_malloc0 (len);
   if (sysctl (mib, G_N_ELEMENTS (mib), data_buf, &len, nullptr, 0) == -1)
       return TRUE;
   data = ((char**)data_buf)[0];
@@ -2330,3 +2609,35 @@ terminal_screen_paste_text (TerminalScreen* screen,
 
   vte_terminal_paste_text (VTE_TERMINAL (screen), text);
 }
+
+GIcon*
+terminal_screen_get_window_icon(TerminalScreen* screen)
+{
+  g_return_val_if_fail(TERMINAL_IS_SCREEN(screen), nullptr);
+
+  auto const priv = screen->priv;
+
+  if (priv->icon_image)
+    return priv->icon_image;
+  else if (priv->icon_color)
+    return priv->icon_color;
+  else
+    return nullptr;
+}
+
+GIcon*
+terminal_screen_get_icon(TerminalScreen* screen)
+{
+  g_return_val_if_fail(TERMINAL_IS_SCREEN(screen), nullptr);
+
+  auto const priv = screen->priv;
+
+  if (priv->has_progress)
+    return terminal_screen_ensure_icon_progress(screen);
+  else if (priv->icon_image)
+    return priv->icon_image;
+  else if (priv->icon_color)
+    return priv->icon_color;
+  else
+    return nullptr;
+}
