@@ -776,7 +776,7 @@ sequences = [
             comment='pcterm'),
     seq_CSI('XTERM_RPM', 'r', pintro=(ParameterIntro.WHAT,),
             comment='xterm restore DEC private mode'),
-    seq_CSI('DECSLRM', 's', flags=Flags.NOP,
+    seq_CSI('DECSLRM', 's',
             comment='set left and right margins'),
     seq_CSI('SCOSC', 's',
             comment='SCO save cursor'),
@@ -922,7 +922,7 @@ sequences = [
             comment='restore terminal state'),
     seq_DCS('XTERM_STCAP', 'p', intermediates=(Intermediate.PLUS,), flags=Flags.NOP,
             comment='xterm set termcap/terminfo'),
-    seq_DCS('DECSIXEL', 'q', flags=Flags.UNRIPE | Flags.HANDLER_RV,
+    seq_DCS('DECSIXEL', 'q', flags=Flags.NOP,
             comment='SIXEL graphics'),
     seq_DCS('DECRQSS', 'q', intermediates=(Intermediate.CASH,),
             comment='request selection or setting'),
